package Data::Filter::Factory;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Data::Filter::Factory - Factory for creating filter rules according to the given regular grammar

=head1 SYNOPSIS

 use Data::Filter::Factory;

 # Create filter factory object
 my $factory = Data::Filter::Factory->new_from_file($grammar_file);

 # Create filter by parsing rule string
 my $filter = $factory->filter_from_string('ID == "abc123456789" and Node/Name eq "d1c2b3a4"');

=head1 DESCRIPTION

This factory parses the regular grammar from given string or file and then it can be used to
construct object filters by parsing given strings containing filter rules.

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1), Parse::RecDescent(3), Data::Filter::Rule(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #--- DEVELOPMENT ONLY ---#
#use Smart::Comments; #--- DEVELOPMENT ONLY ---#

#-- Perl CPAN modules ---------------------------------------------------------#
use Parse::RecDescent;
use JSON;

#-- Custom application modules ------------------------------------------------#
use Data::Filter::Rule::Constant;
use Data::Filter::Rule::List;
use Data::Filter::Rule::Variable;

use Data::Filter::Rule::Existence;

use Data::Filter::Rule::OperationUnary;
use Data::Filter::Rule::OperationBinary;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Translation table from operation type to operation class
my %TYPE2CLASS = (
        'const' => 'Data::Filter::Rule::Constant',
        'list'  => 'Data::Filter::Rule::List',
        'var'   => 'Data::Filter::Rule::Variable',

        'exists' => 'Data::Filter::Rule::Existence',
        'EXISTS' => 'Data::Filter::Rule::Existence',

        'not'   => 'Data::Filter::Rule::OperationUnary',
        'NOT'   => 'Data::Filter::Rule::OperationUnary',
        'and'   => 'Data::Filter::Rule::OperationBinary',
        'AND'   => 'Data::Filter::Rule::OperationBinary',
        'or'    => 'Data::Filter::Rule::OperationBinary',
        'OR'    => 'Data::Filter::Rule::OperationBinary',
        'xor'   => 'Data::Filter::Rule::OperationBinary',
        'XOR'   => 'Data::Filter::Rule::OperationBinary',

        'eq'    => 'Data::Filter::Rule::OperationBinary',
        'EQ'    => 'Data::Filter::Rule::OperationBinary',
        '=='    => 'Data::Filter::Rule::OperationBinary',
        'ne'    => 'Data::Filter::Rule::OperationBinary',
        'NE'    => 'Data::Filter::Rule::OperationBinary',
        '!='    => 'Data::Filter::Rule::OperationBinary',
        'gt'    => 'Data::Filter::Rule::OperationBinary',
        'GT'    => 'Data::Filter::Rule::OperationBinary',
        '>'     => 'Data::Filter::Rule::OperationBinary',
        'lt'    => 'Data::Filter::Rule::OperationBinary',
        'LT'    => 'Data::Filter::Rule::OperationBinary',
        '<'     => 'Data::Filter::Rule::OperationBinary',
        'ge'    => 'Data::Filter::Rule::OperationBinary',
        'GE'    => 'Data::Filter::Rule::OperationBinary',
        '>='    => 'Data::Filter::Rule::OperationBinary',
        'le'    => 'Data::Filter::Rule::OperationBinary',
        'LE'    => 'Data::Filter::Rule::OperationBinary',
        '<='    => 'Data::Filter::Rule::OperationBinary',

        'like'  => 'Data::Filter::Rule::OperationBinary',
        'LIKE'  => 'Data::Filter::Rule::OperationBinary',
        '=~'    => 'Data::Filter::Rule::OperationBinary',

        'in'    => 'Data::Filter::Rule::OperationBinary',
        'IN'    => 'Data::Filter::Rule::OperationBinary',
        '~~'    => 'Data::Filter::Rule::OperationBinary',

        '+'     => 'Data::Filter::Rule::OperationBinary',
        '-'     => 'Data::Filter::Rule::OperationBinary',
        '*'     => 'Data::Filter::Rule::OperationBinary',
        '/'     => 'Data::Filter::Rule::OperationBinary',
        '%'     => 'Data::Filter::Rule::OperationBinary',
        '.'     => 'Data::Filter::Rule::OperationBinary',
    );

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;

    # Parse::RecDescent settings
    $::RD_ERRORS = 1; # Make sure the parser dies when it encounters an error
    #$::RD_WARN   = 1; # Enable warnings. This will warn on unused rules &c.
    #$::RD_HINT   = 1; # Give out hints to help fix problems.
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Data::Filter::Factory->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Data::Filter::Factory reference
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method
  See Also  : _init() method

=cut

sub new()
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item new_from_file($) [PUBLIC,STATIC]

  Usage     : my $instance = Data::Filter::Factory->new_from_file($filename)
  Purpose   : Create and return reference to the new instance
  Arguments : string $grammar_file - name of the grammar file
  Returns   : Data::Filter::Factory reference
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses new() method
  See Also  : new() method

=cut

sub new_from_file($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($grammar_file,) = @_;
    confess "Filter grammar file name must be given as argument" unless $grammar_file;
    confess "Filter grammar file '$grammar_file' is not ordinary file" unless -f $grammar_file;

    # Load grammar from external file to string and untaint it
    open (my $gf, $grammar_file) or die "Can`t open filter grammar file '$grammar_file'\n";
    my $grammar = join("", <$gf>);
    close($gf);
    if ($grammar =~ /^(.+)$/s) { $grammar = $1; }

    # Create the factory object based on the loaded grammar
    return $class->new($grammar);
}

=item filter_from_string($) [PUBLIC]

  Usage     : my $filter = $factory->filter_from_string($filter_rule);
  Purpose   : Create filter object by parsing given rule string
  Arguments : string $filter_rule - filter rule to parse
  Returns   : Data::Filter::Rule filter
  Throws    : Croaks, if invoked on class

=cut

sub filter_from_string($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter_rule, $compile) = @_;
    confess "Filter rule must be given as argument" unless $filter_rule;

    my $rule_tree;
    eval {
        $rule_tree = $self->{PARSER}->start($filter_rule);
    };
    if ($@) { die "Syntax error in filter rule '$filter_rule': $@"; }

    return $rule_tree;
}

=item rule($;@) [PUBLIC,STATIC]

  Usage     : my $instance = Data::Filter::Rule->new_rule($rule_class)
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Data::Filter::Rule filter
  Throws    : Croaks, if invoked on object

=cut

sub rule($;@)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($type, @args) = @_;
    confess "Type of the filter rule must be given" unless $type;
    confess "Unknown filter rule type '$type'" unless $TYPE2CLASS{$type};

    my $rule_class = $TYPE2CLASS{$type};
    return $rule_class->new($type, @args);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init($) [PROTECTED]
#
#  Usage     : In constructor: return $self->_init(@_);
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _init($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($grammar,) = @_;
    confess "Filter grammar must be given as argument" unless $grammar;

    # Create inner RecDescent parser object
    $self->{PARSER} = Parse::RecDescent->new($grammar);
    confess "Error occured during parsing the filter grammar" unless $self->{PARSER};

    return $self;
}

=pod

=back

=cut

1;
