package Mentat::Negistry;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Negistry - Short module description

=head1 SYNOPSIS

 use Mentat::Negistry;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Parent::Module(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use JSON;
use MongoDB::BSON::Binary;

#-- Custom application modules ------------------------------------------------#
use Mentat::Core;
use Value::Convertor;
use Mentat::Storage::Mongo;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant SOURCE_NEGISTRY => 'negistry';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Mentat::Negistry->new();
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Mentat::Negistry reference on success, croaks on failure
  Throws    : Dies, if not invoked on class
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    ### Creating Mentat.Negistry instance...
    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Destroying Mentat.Negistry instance...
}

=item get_current_groups() [PUBLIC]

  Usage     :
  Purpose   : Fetch list of currently defined groups from database
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on class

=cut

sub get_current_groups()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my ($groups, $count_all) = $self->{STORAGE}->find();
    my $result = {};

    # Translate result into hash to allow easy searching
    foreach my $g (@$groups) {
        $result->{$g->{_id}} = $g;
    }

    return $result;
}

=item load_groups() [PUBLIC]

  Usage     :
  Purpose   : Load group from external JSON dump file
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on class

=cut

sub load_groups()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $data = $self->_load_json_dump();
    my %ids  = ();

    my $result = {};
    my ($resolved_abuses, $net, $id, $desc);
    while (my ($network, $d) = each (%$data)) {
        # Skip some special sections within the dump file
        next if ($network =~ /^_/);

        # Get list of all known abuse contacts for this network
        $resolved_abuses = $d->{resolved_abuses};

        # Prepare network record for each abuse group
        foreach my $abuse (@$resolved_abuses) {
            print STDERR "Network: $network, Abuse: $abuse\n" if $DEVEL and $abuse eq 'abuse@cesnet.cz';
            # Generate unique identifier
            while (1) {
                $id = Mentat::Core::uid();
                unless ($ids{$id}) { $ids{$id}++; last; }
            }
            $desc = ($d->{descr}) ? join(', ', @{$d->{descr}}) : 'Unknown';
            # Prepare the network record
            $net = {
                '_id'         => $id,
                'network'     => $network,
                'netname'     => $d->{netname},
                'first'       => $d->{first},
                'last'        => $d->{last},
                'description' => $desc,
                'source'      => SOURCE_NEGISTRY(),
            };
            $net->{network} =~ s/[ ]//g;
            if (defined $d->{ip4_start}) {
                $net->{ip4_start} = $d->{ip4_start};
                $net->{ip4_end}   = $d->{ip4_end};
                $net->{type}      = 'ipv4';
            }
            else {
                $net->{ip6_addr}   = $d->{ip6_addr};
                $net->{ip6_prefix} = $d->{ip6_prefix};
                $net->{type}      = 'ipv6';
            }
            push(@{$result->{$abuse}}, $net);
        }
    }

    # Sort networks for every abuse group
    while (my ($abuse, $networks) = each (%$result)) {
        #print STDERR "Sorting abuse: $abuse, networks: ".Dumper($networks)."\n" if $DEVEL and $abuse eq 'abuse@cesnet.cz';
        my @sorted = $self->_sort_nets(@$networks);
        $result->{$abuse} = \@sorted;
        #print STDERR "Sorted abuse: $abuse, networks: ".Dumper($result->{$abuse})."\n" if $DEVEL and $abuse eq 'abuse@cesnet.cz';
    }

    print STDERR Dumper($result) if $DEVEL;
    return $result;
}

=item insert_group($$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on class

=cut

sub insert_group($$) {
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($id, $networks) = @_;

    return $self->{STORAGE}->insert({
            '_id'          => $id,
            'ts'           => time(),
            'managers'     => [],
            'source'       => SOURCE_NEGISTRY(),
            'subnet_cache' => 0,
            'networks'     => $networks,
            'rep_mode'     => 'summary',
            'rep_emails'   => [$id],
            'rep_filters'  => [],
            'rep_mute'     => 0,
            'rep_redirect' => 1,
        });
}

=item update_group($$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on class

=cut

sub update_group($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($id, $current, $wanted) = @_;

    if ($DEVEL and $id eq 'abuse@cesnet.cz') {
        Log::Loger->notice('test', "Abuse: $id, Current: ".Dumper($current)."\nWanted: ".Dumper($wanted));
    }

    # Convert to hash to allow searching
    my %cur;
    foreach my $c (@$current) {
        next unless ref $c;
        $cur{$c->{network}} = $c;
    }
    my %wan;
    foreach my $w (@$wanted) {
        next unless ref $w;
        $wan{$w->{network}} = $w;
    }

    my %new; my $nid; my $net; my $changes = 0; my %nets = ();
    # Walk through all currently defined networks, remove the deleted ones
    foreach my $c (values %cur) {
        $nid = $c->{_id};
        $net = $c->{network};

        # Network ID must be unique
        while ($new{$nid}) {
            $nid = Mentat::Core::uid();
            $c->{_id} = $nid;
        }

        # Network must be unique
        next if $nets{$net};

        # Ignore manually created network records = put them back
        if ($c->{source} ne SOURCE_NEGISTRY()) {
            $new{$nid} = $c;
            $nets{$net}++;
        }
        else {
            # Current network stil exists, copy it
            if ($wan{$net}) {
                $new{$nid} = $c;
                $nets{$net}++;
            }
            # Network was removed, leave it alone
            else {
                $changes++;
            }
        }
    }
    # Walk through all wanted networks, insert missing ones
    foreach my $w (values %wan) {
        $nid = $w->{_id};
        $net = $w->{network};

        # Network ID must be unique
        while ($new{$nid}) {
            $nid = Mentat::Core::uid();
            $w->{_id} = $nid;
        }

        # Network must be unique
        next if $nets{$net};

        # Network is newly added
        unless ($cur{$net}) {
            $new{$nid} = $w;
            $nets{$net}++;
            $changes++;
        }
    }

    # Proceed only if the network set changed somehow
    if ($changes) {
        # Sort the networks
        my @nets = $self->_sort_nets(values(%new));

        # Update the record within the storage
        my $result = $self->{STORAGE}->update(
            {
                '_id'    => $id,
                'source' => SOURCE_NEGISTRY(),
            },
            {
                '$set' => {
                'source'   => SOURCE_NEGISTRY(),
                'networks' => \@nets,
            },
        });

        return $result->{n};
    }
    return undef;
}

=item remove_group($) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on class

=cut

sub remove_group($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($id,) = @_;

    my $result = $self->{STORAGE}->remove({
        '_id' => $id,
        #'source' => SOURCE_NEGISTRY()
    });

    return $result->{n};
}

=item get_networks() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on class

=cut

sub get_networks()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my ($groups, $count_all) = $self->{STORAGE}->find({
        'subnet_cache' => 1,
    });

    my $result = ();
    foreach my $g (@$groups) {
        my $subnets = $g->{'networks'};
        foreach my $s (@$subnets) {
            push (@$result, $s->{network});
        }
    }

    return $self->_uniq(@$result);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Negistry instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : Mentat::Negistry $self reference
#  Throws    : Dies, if not invoked on instance or given invalid arguments
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($db, $col, $whois_dump_file) = @_;

    $self->{STORAGE}    = Mentat::Storage::Mongo->new(database => $db, collection => $col);
    $self->{WHOIS_DUMP} = $whois_dump_file;

    return $self;
}

sub _load_json_dump()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $json = JSON->new;
    my $filename = $self->{WHOIS_DUMP};

    my $json_text = do {
       open(my $json_fh, "<:encoding(UTF-8)", $filename) or die("Can't open \$filename\": $!\n");
       local $/;
       <$json_fh>
    };

    return $json->decode($json_text);
}

sub _sort_nets(@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    return sort { $b->{'first'} <=> $a->{'first'} }
           sort { $a->{'last'} <=> $b->{'last'} } @_;
}

sub _uniq(@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my %seen;
    return grep { !$seen{$_}++ } @_;
}

=pod

=back

=cut

1;
