package Mentat::Storage::Mongo;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Storage::Mongo - Mentat object storage in MongoDB database

=head1 SYNOPSIS

 #*******************************************************************************##
 # !!! MORE USAGE EXAMPLES MAY BE FOUND IN Mentat.Storage.Mongo.t TEST SCRIPT !!! #
 #*******************************************************************************##

 use Mentat::Storage::Mongo;

 # Instantinate the database engine
 my $engine = Mentat::Storage::Mongo->new(database => 'db', collection => 'coll');
 my $engine = Mentat::Storage::Mongo->new(connection => [host => 'host', port => 1234, timeout => 2000], database => 'db', collection => 'coll');

 # Any of the following methods may croak, if the database error is encountered,
 # so wrap everything in eval
 eval {

    # Insert new record into database
    my $record_id = $instance->insert({a => 1, b => 1});

    # Fetch the object back from database
    my $object = $engine->fetch($record_id);

    # Delete the single object from database (based on unique OID)
    my $rv = $engine->delete($record_id);


    # Find all objects in database
    my ($objects, $count) = $engine->find();

    # Find and filter
    my ($objects, $count) = $engine->find({'_id' => $record_id1});
    my ($objects, $count) = $engine->find({'a' => {'$gt' => 4}});

    # TODO: Projection is broken and doesn`t work for some reason

    # Find, filter and sort
    my ($objects, $count) = $engine->find({'a' => {'$gt' => 4}}, undef, {'b' => -1});

    # Find, filter, sort and limit
    my ($objects, $count) = $engine->find({'a' => {'$gt' => 4}}, undef, {'b' => -1},5);

    # Find, filter, sort and skip
    my ($objects, $count) = $engine->find({'a' => {'$gt' => 4}}, undef, {'b' => -1}, undef, 10);

    # Find, filter, sort, limit and skip
    my ($objects, $count) = $engine->find({'a' => {'$gt' => 4}}, undef, {'b' => -1}, 5, 10);


    # Using iteration on cursor, arguments are the same as with find() method
    my @result = [];
    my ($cursor, $cnta) = $engine->find_i();
    # !!!PREFERED METHOD!!! Now either use internal cursor iteration with optional limit
    # This method uses the closures @result is visible in callback method
    sub cb($) {
        push(@result, @_); # OR DO SOME OTHER STUFF HERE
    }
    my $cntb = $engine->cursor_iterate($cursor, \&cb, $limit);
    # !!!DISCOURAGED METHOD!!! Or use cursor directly
    while (my $doc = $cursor->next) {
        push(@result, $doc);  # OR DO SOME OTHER STUFF HERE
    }
    # @result now contains all data


    # Count number of all records in database
    my $count = $engine->count();

    # Count number of records in database, which pass the given filter condition
    my $count = $engine->count({'_id' => $record_id});
    my $count = $engine->count({'a' => {'$gt' => 4}});


    # Update single or multiple records
    my $rv = $instance->update({'_id' => $record_id1},{a => 'x', b => 'y'});
    my $rv = $instance->update({a => {'$lte'=>2}},{'$inc' => {b => 3 }});
    my $rv = $instance->update({a => {'$lte'=>2}},{'$inc' => {b => 3}},{multiple => 1});


    # Remove all Mentat::Messages in database
    my $rv = $engine->remove();

    # Remove all Mentat::Messages in database, which pass the given filter condition
    my $rv = $engine->remove({'_id' => $record_id});
    my $rv = $engine->remove({'a' => {'$gt' => 4}});


    # After find(), count() and remove() methods you may call get_last_filter()
    # method to determine the last filter-sort-limit-skip condition
    my $filter = $engine->get_last_filter();

 };
 if($@) {
    print STDERR "MongoDB error: $@\n";
 }

 #*******************************************************************************##
 # !!! MORE USAGE EXAMPLES MAY BE FOUND IN Mentat.Storage.Mongo.t TEST SCRIPT !!! #
 #*******************************************************************************##

=head1 DESCRIPTION

This engine is used for permanent storage of Mentat objects in document oriented
MongoDB database. It is the implementation of the base Mentat::Storage
interface. Implementation is currently based on MongoDB module, which is
available in CPAN repository.

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), MongoDB(3), Mentat::Storage(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use DateTime;
use Time::HiRes qw(time);
use Clone qw(clone);
use Tie::IxHash;

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use JSON;
use MongoDB;
use MongoDB::BSON::Binary;

#-- Custom application modules ------------------------------------------------#
use Mentat::Storage;
use Value::Convertor;

# Modules from Cesnet Toolkit
use Log::Loger;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

use constant DFLT_LIMIT   => 500000;
#use constant DFLT_TIMEOUT => 200000;
# RO: Nez vyresim problem s mongodb, potrebuji vyssi limit:
use constant DFLT_TIMEOUT => 1000000;

use constant KEY_ID => '_id';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.6';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Storage');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

sub DESTROY
{
    my $self    = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    $self->disconnect();
}

=item insert($) [PUBLIC]

  Usage     : my ($record_id, $error) = $storage->insert($object);
  Purpose   : Insert the given object into storage
  Arguments : MIXED $object - object to be stored into storage
  Returns   : STRING $oid - ID of the inserted object on SUCCESS, undef on FAILURE
              STRING $error - undef on SUCCESS, error string on failure
  Throws    : Dies, if not invoked on instance
              Croaks on insert failure
  Comment   : This method dies in case of insert failure, it must be taken care
              of on higher levels !
  Comment   : This method attempts to automatically convert given object to hash,
              either using internal convertor (if given upon instantination), or
              using object`s to_hash() method (if defined)!

=cut

sub insert($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($object,) = @_;

    unless (ref $object eq 'HASH') {
        if ($self->_convertor) {
            $object = $self->_convertor->convert($object);
        }
        elsif (blessed($object) and $object->can('to_hash')) {
            $object = $object->to_hash();
        }
    }

    # Try to insert the object into MongoDB
    #Log::Loger->debug(__PACKAGE__, "[MSM] Inserting object: ".Dumper($object));
    return $self->{STORAGE}->insert($object, {safe => 1});
}

=item fetch($) [PUBLIC]

  Usage     : my $object = $storage->fetch($oid);
  Purpose   : Fetch the Mentat message with given ID from storage
  Arguments : STRING $oid - unique identifier of the object in storage
  Returns   : HASH REFERENCE $object reference on SUCCESS,
              undef if object with given ID was not found,
  Throws    : Dies, if not invoked on instance
              Croaks on query failure
  Comments  :
  See Also  :

=cut

sub fetch($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($oid,) = @_;

    # Execute query in MongoDB
    my $cursor = $self->{STORAGE}->query({KEY_ID() => $oid})->limit(1);

    return $cursor->next;
}

=item ref_fetch($) [PUBLIC]

  Usage     : my $object = $storage->ref_fetch($ref);
  Purpose   : Fetch the object with given reference
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance
              Croaks on query failure

=cut

sub ref_fetch($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($ref,) = @_;

    # Execute query in MongoDB
    my $s = $self->{DATABASE}->get_collection($ref->{ref});
    my $cursor = $s->query({KEY_ID() => $ref->{id}})->limit(1);

    return $cursor->next;
}

=item update($$;$) [PUBLIC]

  Usage     : my $result = $storage->update($filter, $values, $options);
  Purpose   : Update the existing object in the storage
  Arguments : HASH REFERENCE $filter - Update filter [MANDATORY]
              HASH REFERENCE $values - Object values [MANDATORY]
              HASH REFERENCE $options - Update options [OPTIONAL]
  Returns   : HASH REFERENCE containing information about the operation result:
                 ok  => success flag
                 n   => number of affected items
                 err => error flag
  Throws    : Dies, if not invoked on instance

=cut

sub update($$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $values, $options) = @_;

    # Modify options so that the update is safe
    $options = ($options && ref $options eq 'HASH')?$options:{};
    $options->{'safe'} = 1 unless exists $options->{'safe'};

    return $self->{STORAGE}->update($filter, $values, $options);
}

=item delete($) [PUBLIC]

  Usage     : my $result = $storage->delete($oid);
  Purpose   : Delete the object with given ID from the storage
  Arguments : STRING $oid - unique identifier of the object in storage
  Returns   : HASH REFERENCE containing information about the operation result:
                 ok  => success flag
                 n   => number of affected items
                 err => error flag
  Throws    : Dies, if not invoked on instance
  Comments  : This operation should always remove only one object
  See Also  : For bulk delete operation see remove() method

=cut

sub delete($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($oid,) = @_;

    my $filter = { KEY_ID() => $oid };
    $self->_set_last_filter('DELETE', $filter);
    Log::Loger->warn(__PACKAGE__, "[MSM] Executing MongoDB command: ".$self->get_last_filter());

    return $self->{STORAGE}->remove($filter, {safe => 1});
}

=item count(;$) [PUBLIC]

  Usage     : my $count = $storage->count(); or
              my $count = $storage->count($filter);
  Purpose   : Get the number of Mentat messages in storage
  Arguments : HASH REFERENCE $filter - Query filter [OPTIONAL]
  Returns   : INTEGER on SUCCESS
  Throws    : Dies, if not invoked on instance
              Dies on operation failure

=cut

sub count(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter,) = @_;

    $self->_set_last_filter('COUNT', $filter);
    Log::Loger->warn(__PACKAGE__, "[MSM] Executing MongoDB command: ".$self->get_last_filter());

    # Unless the filter was explicitly specified, use shortcut
    unless ($filter) {
        return $self->{STORAGE}->count();
    }
    # Otherwise apply the filter first
    else {
        return $self->{STORAGE}->query($filter)->count();
    }
}

=item find_i(;$$$$$) [PUBLIC]

  Usage     : my ($objects, $count) = $storage->find_i(); or
              my ($objects, $count) = $storage->find_i($filter);
  Purpose   : Find the object(s) according to the given filter and return the cursor to the result
  Arguments : HASH REFERENCE $filter - Query filter [OPTIONAL]
              HASH REFERENCE $selection - Record attribute selection [OPTIONAL]
              HASH REFERENCE $sort - Sorting specification [OPTIONAL]
              INTEGER $limit - Maximum number of records to be returned [OPTIONAL]
              INTEGER $skip - Number of objects to skip from the beginning of the result set [OPTIONAL]
              BOOL $quick - Quick search, do not perform any development tests [OPTIONAL]
  Returns   : ARRAY REFERENCE containing objects
              INTEGER total count of objects
  Throws    : Dies, if not invoked on instance

=cut

sub find_i(;$$$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $projection, $sort, $limit, $skip, $quick) = @_;
    confess "Received invalid filter as argument" if $filter and ref $filter ne 'HASH';

    $limit = ($limit) ? int($limit) : 0;
    $skip  = ($skip) ? int($skip) : 0;

    $self->_set_last_filter('FIND', $filter, $projection, $sort, $limit, $skip, $quick);
    Log::Loger->warn(__PACKAGE__, "[MSM] Executing MongoDB command: ".$self->get_last_filter());

    # Execute the query
    my $cursor = $self->{STORAGE}->find(Tie::IxHash->new(%$filter));
    #my $cursor = $self->{STORAGE}->find($filter);

    # Count the results
    my $count = undef;
    Log::Loger->debug(__PACKAGE__, "[MSM] Counting all results") if $DEVEL and not $quick;
    $count = $cursor->count() unless $quick;

    # Apply the projection
    if ($projection) {
        Log::Loger->debug(__PACKAGE__, "[MSM] Projecting results") if $DEVEL;
        $cursor = $cursor->fields($projection);
    }

    # Apply the sorting
    if ($sort) {
        Log::Loger->debug(__PACKAGE__, "[MSM] Sorting results") if $DEVEL;
        $cursor = $cursor->sort($sort);
    }

    # Skip few first records
    Log::Loger->debug(__PACKAGE__, "[MSM] Skiping '$skip' results from beginning") if $DEVEL;
    $cursor = $cursor->skip($skip) if ($skip > 0);

    # Return limited number of records
    if ($limit > 0) {
        Log::Loger->debug(__PACKAGE__, "[MSM] Limiting results to max '$limit'") if $DEVEL;
        $cursor = $cursor->limit($limit);
    }
    else {
        Log::Loger->debug(__PACKAGE__, "[MSM] Results are unlimited") if $DEVEL;
    }

    # Info about mongodb index
    $self->_set_last_index($cursor) unless $quick;
    Log::Loger->debug(__PACKAGE__, "[MSM] MongoDB index: ".$self->get_last_index()) if $DEVEL and not $quick;

    # Return the result cursor and count
    return ($cursor, $count);
}

=item find(;$$$$$) [PUBLIC]

  Usage     : See the find_i() method for details
  Purpose   : Find the object(s) according to the given filter and return them in array
  Arguments : See the find_i() method for details
  Returns   : See the find_i() method for details
  Throws    : Dies, if not invoked on instance

=cut

sub find(;$$$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $projection, $sort, $limit, $skip, $quick) = @_;
    confess "Received invalid filter as argument" if $filter and ref $filter ne 'HASH';

    # Setup some limit, when we are returning all items from database
    $limit = DFLT_LIMIT unless ($limit and (int($limit) > 0));

    # Delegate the work to the find() method
    my ($cursor, $count) = $self->find_i($filter, $projection, $sort, $limit, $skip, $quick);

    # Put all results into the target array
    my @array = ();
    while (my $doc = $cursor->next) {
        push (@array, $doc);
    }

    return (\@array, $count);
}

=item remove(;$) [PUBLIC]

  Usage     : my $result = $storage->remove(); or
              my $result = $storage->remove($filter);
  Purpose   : Remove the objects(s) according to the given filter
  Arguments : HASH REFERENCE $filter - Filter for selecting items to be removed [OPTIONAL]
  Returns   : HASH REFERENCE containing information about the operation result:
                 ok  => success flag
                 n   => number of affected items
                 err => error flag
  Throws    : Dies, if not invoked on instance
              Croaks on query failure
  Comments  : This operation performs bulk delete operation
  See Also  : For single delete operation see delete() method
  Warning   : DELETES ALL DATA, IF GIVEN NO FILTER !!!

=cut

sub remove(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter,) = @_;

    $self->_set_last_filter('REMOVE', $filter);
    Log::Loger->warn(__PACKAGE__, "[MSM] Executing MongoDB command: ".$self->get_last_filter());

    return $self->{STORAGE}->remove($filter, {safe => 1});
}

=item aggregate [PUBLIC]

  Usage     : my $result = $storage->aggregate(@_);
  Purpose   : Call the aggregate operation on internal storage
  Arguments : Same as MongoDB::Collection::aggregate() method
  Returns   : Same as MongoDB::Collection::aggregate() method
  Throws    : Dies, if invoked on class
  See Also  : MongoDB::Collection::aggregate() method for more details

=cut

sub aggregate
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    Log::Loger->warn(__PACKAGE__, "[MSM] Executing MongoDB aggregate: ".Dumper (@_));
    return $self->{STORAGE}->aggregate(@_);
}

=item distinct($;$$) [PUBLIC]

  Usage     : my $result = $storage->distinct($key, $collection, $query);
  Purpose   : Call the distinct operation on internal storage
  Arguments : Same as MongoDB::Collection::distinct() method
  Returns   : Same as MongoDB::Collection::distinct() method
  Throws    : Dies, if invoked on class
  See Also  : MongoDB::Collection::distinct() method for more details

=cut

sub distinct($;$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($key, $collection, $query) = @_;
    $collection = $self->{CFG_COLLECTION} unless $collection;
    $query = {} unless $query and ref $query eq 'HASH';

    my $r = $self->run_command([
                "distinct" => $collection,
                "key"      => $key,
                "query"    => $query,
            ]);
    return $r->{'values'};
}

=item run_command [PUBLIC]

  Usage     : my $result = $storage->run_command(@_);
  Purpose   : Call the specific operation on internal storage
  Arguments : Same as MongoDB::Database::run_command() method
  Returns   : Same as MongoDB::Database::run_command() method
  Throws    : Dies, if invoked on class
  See Also  : MongoDB::Database::run_command() method for more details

=cut

sub run_command
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    Log::Loger->warn(__PACKAGE__, "[MSM] Executing MongoDB run_command: ".Dumper (@_));
    return $self->{DATABASE}->run_command(@_);
}

=item disconnect() [PUBLIC,ABSTRACT]

  Usage     : $storage->disconnect();
  Purpose   : Close existing storage connection
  Arguments : NONE
  Returns   : UNDEF on SUCCESS, STRING $error on FAILURE
  Throws    : Dies, if not invoked on instance

=cut

sub disconnect()
{
    my $self    = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    if ($self->{CONNECTION}) {
        $self->{STORAGE}    = undef;
        $self->{DATABASE}   = undef;
        $self->{CONNECTION} = undef;

        return undef;
    }
    else {
        return "Connection was already closed";
    }
}

=item reconnect() [PUBLIC]

  Usage     : $storage->reconnect();
  Purpose   : Optionally close existing storage connection and reopen it
  Arguments : NONE
  Returns   : UNDEF on SUCCESS, STRING $error on FAILURE
  Throws    : Dies, if not invoked on instance

=cut

sub reconnect()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Disconnect first, if connection already exists
    $self->disconnect() if ($self->{CONNECTION});

    # Open new connection
    my $options = {'w' => 1, 'socket_timeout_ms' => $self->{CFG_QUERY_TIMEOUT}};
    if ($self->{CFG_CONNECTION}) {
        my %c = (@{$self->{CFG_CONNECTION}});
        foreach my $k (keys %c) { $options->{$k} = $c{$k}; }
    }
    Log::Loger->notice(__PACKAGE__, "[MSM] (Re)Opening database connection with options:".Dumper($options));
    $self->{CONNECTION} = MongoDB->connect('mongodb://localhost', $options);

    # Select database
    $self->{DATABASE} = $self->{CONNECTION}->get_database($self->{CFG_DATABASE});

    # Connect to speciffic collection
    $self->{STORAGE}  = $self->{DATABASE}->get_collection($self->{CFG_COLLECTION});

    #print STDERR "Connected to '$self->{CFG_DATABASE}':'$self->{CFG_COLLECTION}' with timeout '$self->{CFG_QUERY_TIMEOUT}'\n";

    return undef;
}

=item get_last_filter() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance
  Comments  :
  See Also  :

=cut

sub get_last_filter()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $sort, $limit, $skip) = @_;

    return $self->{LAST_FILTER};
}

=item get_last_index(;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments : INTEGER $verbose - [OPTIONAL]
  Returns   :
  Throws    : Dies, if not invoked on instance
  Comments  :
  See Also  :

=cut

sub get_last_index(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($verbose) = @_;

    return ($verbose)?$self->_encode2json($self->{LAST_INDEX}):$self->{LAST_INDEX}->{cursor};
}

=item cursor_iterate($$;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments : MongoDB::Cursor $cursor - [MANDATORY]
              CODEREF $cb - [MANDATORY]
              INTEGER $limit - [OPTIONAL]
  Returns   :
  Throws    : Dies, if not invoked on instance
  Comments  :
  See Also  :

=cut

sub cursor_iterate($$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($cursor, $cb, $limit) = @_;
    confess "Valid database cursor must be given as argument, received: ".blessed($cursor) unless blessed($cursor) and blessed($cursor) eq 'MongoDB::Cursor';
    confess "Valid callback must be given as argument" unless ref $cb and ref $cb eq 'CODE';

    no strict 'refs';
    my $i = 0;
    while (my $doc = $cursor->next) {
        ++$i;
        &{$cb}($doc);
        last if $limit and $i >= $limit;
    }
    return $i;
}

=item list_collections() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance
  Comments  :
  See Also  :

=cut

sub list_collections()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Connect to speciffic collection
    return $self->{DATABASE}->collection_names();
}

=item list_indexes() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance
  Comments  :
  See Also  :

=cut

sub list_indexes()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Connect to speciffic collection
    return $self->{STORAGE}->get_indexes();
}

=item change_collection($) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments : STRING $collection - [MANDATORY]
  Returns   :
  Throws    : Dies, if not invoked on instance
  Comments  :
  See Also  :

=cut

sub change_collection($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($collection,) = @_;

    # Connect to speciffic collection
    $self->{STORAGE}        = $self->{DATABASE}->get_collection($collection);
    $self->{CFG_COLLECTION} = $collection;
}

#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance
#  Comments  :
#  See Also  :

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %params = (@_) ? @_ : ();
    confess "Database name must be given as argument" unless $params{database};
    confess "Collection name must be given as argument" unless $params{collection};

    # Store the connection parameters
    $self->{CFG_CONNECTION}    = $params{connection};
    $self->{CFG_DATABASE}      = $params{database};
    $self->{CFG_COLLECTION}    = $params{collection};
    $self->{CFG_QUERY_TIMEOUT} = $params{query_timeout} || DFLT_TIMEOUT();

    # Register the optional convertor
    $self->_convertor($params{convertor});

    # Open the connection
    $self->reconnect();

    return $self;
}

# _create_indexes(@) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance
#  Comments  :
#  See Also  :

sub _create_indexes(@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my @indexes = @_;

    foreach my $i (@indexes) {
        $self->{STORAGE}->ensure_index(@$i);
    }
}

# _convertor(;$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance
#  Comments  :
#  See Also  :

sub _convertor(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($convertor,) = @_;

    if ($convertor) {
        if (blessed($convertor) and $convertor->can('convert')) {
            $self->{CONVERTOR} = $convertor;
        }
        else {
            confess "Received invalid database object convertor";
        }
    }
    return $self->{CONVERTOR};
}

# _convertor_inv(;$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance
#  Comments  :
#  See Also  :

sub _convertor_inv(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($convertor,) = @_;

    if ($convertor) {
        if (blessed($convertor) and $convertor->can('convert')) {
            $self->{CONVERTOR_INV} = $convertor;
        }
        else {
            confess "Received invalid database object inverse convertor";
        }
    }
    return $self->{CONVERTOR_INV};
}

# _set_last_index(;$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance
#  Comments  :
#  See Also  :

sub _set_last_index(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($cursor) = @_;

    $self->{LAST_INDEX} = $cursor->explain;

    return $self;
}

# _set_last_filter(;$$$$$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance
#  Comments  :
#  See Also  :

sub _set_last_filter(;$$$$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($type, $filter, $projection, $sort, $limit, $skip, $quick) = @_;

    $self->{LAST_FILTER} = 'ERROR';

    my $fstr = $self->_encode2json($filter);
    my $pstr = $self->_encode2json($projection);
    my $sstr = $self->_encode2json($sort);

    $self->{LAST_FILTER}  = "[$type] Database: '$self->{CFG_DATABASE}:$self->{CFG_COLLECTION}', filter '$fstr'";
    $self->{LAST_FILTER} .= " with projection '$pstr'" if $pstr;
    $self->{LAST_FILTER} .= " with sorting '$sstr'" if $sstr;
    $self->{LAST_FILTER} .= " with limit '$limit'" if $limit;
    $self->{LAST_FILTER} .= " with skip '$skip'" if $skip;
    $self->{LAST_FILTER} .= ", QUICK" if $quick;
    return $self;
}

# _encode2json($) [PROTECTED]
#
#  Usage     : my $json = $self->_encode2json($anything);
#  Purpose   : Encode given object structure to JSON string
#  Arguments : MIXED $anything - any argument [MANDATORY]
#  Returns   : STRING $json
#  Throws    : Dies, if not invoked on instance
#  Comments  : Internally uses _unbless() method to stringify blessed references
#  See Also  : _unbless() method

sub _encode2json($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($hash,) = @_;

    my ($hash_json, $tmp);
    eval {
        $hash_json = ($hash)?encode_json($hash):'';
    };
    if ($@ and $@ =~ /allow_blessed|convert_blessed/) {
        $tmp       = clone($hash);
        $tmp       = $self->_unbless($tmp);
        $hash_json = encode_json($tmp);
    }
    return $hash_json;
}

# _unbless($) [PROTECTED]
#
#  Usage     : my $result = $self->_unbless($anything);
#  Purpose   : Replace all blessed objects in given data structure
#  Arguments : MIXED $anything - any argument [MANDATORY]
#  Returns   : HASHREF $result
#  Throws    : Dies, if not invoked on instance
#  Comments  : VERY UGLY HACK TO ALLOW PRINTING THE RAW MONGODB QUERY, FIXME

sub _unbless($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter,) = @_;

    # Deal with scalar values immediately, but this should not be necessary now
    return $filter unless ref $filter;

    # Alter the values within HASH
    if (ref $filter eq "HASH") {
        my $t;
        foreach my $key (keys(%$filter)) {
            $t = blessed($filter->{$key});
            # MongoDB::BSON::Binary is tricky to display, use HEX for now
            if ($t and ($t eq 'MongoDB::BSON::Binary')) {
                $filter->{$key} = 'HexData(0,"'.Value::Convertor->bin_to_hexstr($filter->{$key}->data).'")';
            }
            # Math:BigInt has method bstr(), use it
            elsif ($t and ($t eq 'Math::BigInt')) {
                $filter->{$key} = 'NumberLong('.$filter->{$key}->bstr().')';
            }
            # This will attempt to stringify any other object
            elsif ($t) {
                $filter->{$key} = "".$filter->{$key};
            }
            # Recursion for HASHES and ARRAYS
            elsif (ref $filter->{$key}) {
                $filter->{$key} = $self->_unbless($filter->{$key});
            }
        }
        return $filter;
    }
    # Alter the values within array
    elsif (ref $filter eq "ARRAY") {
        my $result = [];
        foreach my $key (@$filter) {
            # Shortcut for scalar values
            unless (ref $key) {
                push(@$result, $key);
            }
            # Recursively unbless everything else
            else {
                push(@$result, $self->_unbless($key));
            }
        }
        return $result;
    }
    # Fallback, this should never happen, but attempt to stringify any other object
    else {
        return "$filter";
    }
}

=pod

=back

=cut

1;
